// ---- exec.c ----

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>

#include "syntax.h"
#include "jobs.h"
#include "intern.h"

void convert_word (char *s)
     /* convertit s qui peut contenir des ' " \ 
	et les interpretes
     */
{
  char *r, *w;

  if (*s == '\'') {
    for (r = s + 1, w = s; *r; r++, w++)
      *w = *r;
    *(w-1) = '\0';
  } else if (*s == '"') {
    for (r = s + 1, w=s; *r; r++, w++) {
      if (*r == '\\') {
	r++;
	if (*r == 'n')
	  *w = '\n';
	else 
	  *w = *r;
      } else {
	*w = *r;
      }
    }
    *(w - 1) = '\0';
  } else {
    for (r = s, w = s; *r; r++, w++)
      if (*r == '\\') {
	r++;
	if (*r == 'n')
	  *w = '\n';
	else 
	  *w = *r;
      } else {
	*w = *r;
      }
    *w = '\0';
  }	
}

char **convert (scomm_t scomm, char **input, char **output)
{
  char **argv;
  int i, argc = 1;
  elt_list_t l;

  *input = *output = "";

  l = scomm.elt_list;
  while (l) {
    convert_word (l->elt.word);
    switch (l->elt.type) {
    case ARG: 
      argc++;
      break;
    case IN:
      *input = l->elt.word;
      break;
    case OUT:
      *output = l->elt.word;
      break;
    }
    l = l->next;
  }

  argv = calloc (argc + 1, sizeof *argv);
  convert_word (scomm.name);
  argv[0] = scomm.name;
  argv[argc] = NULL;
  l = scomm.elt_list;
  i = 1;
  while (l) {
    if (l->elt.type == ARG) {
      argv[i] = l->elt.word;
      i++;
    }
    l = l->next;
  }
  return argv;
}

int is_intern (char *name)
{
  return 
    !strcmp (name, "jobs") ||
    !strcmp (name, "wait") ||
    !strcmp (name, "exit") ||
    !strcmp (name, "cd");
}
    

int exec_extern_scomm (scomm_t scomm)
     /* execute une commande simple (externe)
	(attention: ne fait pas de fork)
     */
{
  char **argv, *input, *output;
  int fd_in, fd_out, ret_val;

  argv = convert (scomm, &input, &output);

  if (strcmp (input, "")) {
    if ((fd_in = open (input, O_RDONLY)) < 0) {
      fprintf (stderr, "ddsh: %s: cannot open.\n", input);
      return -1;
    }
    dup2 (fd_in, STDIN_FILENO);
    close (fd_in);
  }

  if (strcmp (output, "")) {
    if ((fd_out = open (output, O_WRONLY | O_CREAT, 0644)) < 0) {
      fprintf (stderr, "ddsh: %s: cannot open.\n", output);
      return -1;
    }
    dup2 (fd_out, STDOUT_FILENO);
    close (fd_out);
  }

  ret_val = execvp (argv[0], argv);
  perror (argv[0]);
  free (argv);
  return abs (ret_val);
}


int exec_if_is_intern (scomm_t scomm, int *ret)
     /* execute une commande interne du shell 
	renvoie -1 si erreur
	sinon renvoie renvoie 1 ssi scomm est une commande
	interne. Le code de retour est mis dans *ret
      */
{
  char **argv, *input_name, *output_name;
  FILE *fin = stdin, *fout = stdout;

  if (!is_intern (scomm.name)) {
    return 0;
  } 

  argv = convert (scomm, &input_name, &output_name);

  if (strcmp (input_name, ""))
    if ((fin = fopen (input_name, "r")) == NULL) {
      perror (input_name);
      free (argv);
      return -1;
    }
  
  if (strcmp (output_name, ""))
    if ((fout = fopen (output_name, "w")) == NULL) {
      perror (output_name);
      free (argv);
      return -1;
    }
  
  if (!strcmp (scomm.name, "jobs")) {
    *ret = dd_jobs (fin, fout, argv);
  } else if (!strcmp (scomm.name, "exit")) {
    *ret = dd_exit (fin, fout, argv);
  } else if (!strcmp (scomm.name, "wait")) {
    *ret = dd_wait (fin, fout, argv);
  } else if (!strcmp (scomm.name, "cd")) {
    *ret = dd_cd (fin, fout, argv);
  }
  if (fin != stdin) fclose (fin);
  if (fout != stdout) fclose (fout);
  free (argv);
  return 1;
}

int wait_pipe (pipe_t p)
{
  int r = 0;
  pipe_t l;

  l = p;
  while (l) {
    if (l->pid > 0)
      if (waitpid (l->pid, &r, 0) < 0)
	perror ("wait_pipe waitpid");
    l = l->next;
  }
  return r;
}

int exec_pipe (pipe_t p, int sync)
     // execute un tube (attention: fait des fork)
{
  pipe_t l;
  int last_out; /* descripteur sur le cot lecture du tube de sortie
		   de la commande prcdente */
  pid_t pid;
  int son_stat, r;

  if (!p) return 0;
  if (!sync && !can_create_new_job()) {
    fprintf (stderr, "ddsh: Sorry, job table is full.\n");
    return 1;
  }

  l = p;
  last_out = dup (STDIN_FILENO);
  while (l->next) {
    if (is_intern (l->scomm.name)) {
      fprintf (stderr, "ddsh: Cannot pipe intern command %s, ignoring.\n",
	       l->scomm.name);
      l->pid = (pid_t) -1;
    } else {
      pipe (l->fd);
      if ((pid = fork ())) { // pere
	dup2 (l->fd[0], last_out);
	close (l->fd[1]);
	close (l->fd[0]);
	l->pid = pid;
      } else { // fils
	dup2 (last_out, STDIN_FILENO);
	close (last_out);
	dup2 (l->fd[1], STDOUT_FILENO);
	close (l->fd[1]);
	close (l->fd[0]);
	exit (exec_extern_scomm (l->scomm));
      }
    }
    l = l->next;
  }
  
  if (sync) {
    r = exec_if_is_intern (l->scomm, &son_stat);
    if (r < 0)
      return 1;
    if (r) {
      l->pid = (pid_t) -1;
      if (put_in_fg_job (make_job (p), &son_stat) < 0)
	fprintf (stderr, "There is already a fg job!!!!\n");
      return son_stat;
    }
  }

  if ((pid = fork ())) { // pere
    close (last_out);
    l->pid = pid;
    if (sync) {
      if (put_in_fg_job (make_job (p), &son_stat) < 0)
	fprintf (stderr, "There is already a fg job!!!!\n");
      return son_stat;
    } else {
      add_job (stdout, make_job (p));
      return 0;
    }    
  } else { // fils
    dup2 (last_out, STDIN_FILENO);
    close (last_out);
    exit (exec_extern_scomm (l->scomm));
  }
  return son_stat;
}

int exec_comm (comm_t comm)
     // execute une commande complete (fait les fork)
{
  comm_t l = comm;
  int ret_val = 0;
  
  while (l) {
    ret_val = exec_pipe (l->pipe, l->sync);
    l = l->next;
  }
  return WEXITSTATUS (ret_val);
}

