// ---- intern.c ----

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <dirent.h>

#include "syntax.h"
#include "jobs.h"
#include "intern.h"

int dd_jobs (FILE *in, FILE *out, char **argv)
{
  update_print_jobs (out);

  return 0;
}


pid_t get_last_pid (job_t job)
{
  if (!job)
    return (pid_t) -2;

  while (job->next)
    job = job->next;
  
  return job->pid;
}

int dd_wait (FILE *in, FILE *out, char **argv)
{
  pid_t pid;
  int id = -1, ret;

  if (argv[1]) { 
    id = atoi (argv[1]);
  } else
    pid = -1;

  if (pid < -1) {
    fprintf (stderr, "ddsh: %d: no such job\n", id); 
    return 1;
  }

  wait_job (id, &ret);
  
  if (id >= 0)
    update_print_if_dead_jobs (out);

  return ret;  
}

int dd_cd (FILE *in, FILE *out, char **argv)
{
  char *pathname;

  if (argv[1])
    pathname = argv[1];
  else
    pathname = getenv ("HOME");

  if (chdir (pathname) < 0) {
    perror ("cd");
    return 1;
  }

  return 0;
}

int dd_exit (FILE *in, FILE *out, char **argv)
{
  int ret_val = 0;
  static int warned = 0;

  if (exists_job() && !warned) {
    fprintf (stderr, "You have runnning jobs.\n");
    warned = 1;
    return 1;
  }

  if (argv[1])
    ret_val = atoi (argv[1]);

  kill_jobs (SIGHUP);
  printf ("May the ddsh be with you.\n");
  exit (ret_val);
}
