// ---- syntax.c ----

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>

#include "syntax.h"
#include "jobs.h"
#include "intern.h"


// --- LIBERATION DE L'ESPACE MEMOIRE

void free_token_list (token_list_t l)
{
  token_list_t p;
  while (l) {
    p = l;
    l = l->next;
    if (!p->used) free (p->token);
    free (p);
  }
}

void free_elt_list (elt_list_t l, free_t type)
{
  elt_list_t p;
  while (l) {
    p = l;
    l = l->next;
    if (type == HARD) free (p->elt.word);
    free (p);
  }
}

void free_scomm (scomm_t scomm, free_t type)
{
  free_elt_list (scomm.elt_list, type);
}

void free_pipe (pipe_t l, free_t type)
{
  pipe_t p;
  while (l) {
    p = l;
    l = l->next;
    free_scomm (p->scomm, type);
    free (p);
  }
}

void _free_comm (comm_t l, free_t type)
{
  comm_t p;
  while (l) {
    p = l;
    l = l->next;
    free_pipe (p->pipe, type);
    free (p);
  }
}

void free_comm (comm_t l)
{
  _free_comm (l, HARD);
}

// LECTURE D'UN FICHIER

int nb_bs_mod_2 (char *s, char *p)
     /* renvoie la parit du nb de \ 
	qui sont avant p (inclus) dans s
     */
{
  int nb = 0;
  
  for (; p >= s; p--)
    if (*p == '\\')
      nb ^= 1;
    else 
      return nb;
  return nb;
}

int read_line (char *s, FILE *f)
     // Lit une ligne dans f et la place dans s
     // suppos de taille >= MAX_LINE_SIZE
{
  char *p;
  int len, r;

  p = s + 2;
  len = 2;


  do {
    len -= 2;
    p -= 2;
    *p = '\0';

    if (!fgets (p, MAX_LINE_SIZE - len, f)) {
      // fin de fichier ou erreur
      *s = '\0';
      return E_FGETS;
    }

    len += strlen (p);
    p = s + len;
    if (*(p - 1) != '\n') { // ligne trop longue
      if (*(p - 1) == '\\') fgetc(f);
      while ((r = fgetc(f)) != EOF &&
	     r != '\n')
	if (r == '\\') fgetc(f);
      *s = '\0';
      return E_TOO_LONG;
    }
  } while (p >= s + 2 && nb_bs_mod_2 (s, p - 2));

  *(p - 1) = '\0';
  return 0;
}



char *take (char **s)
     /* prend un 'token' ("chaine", 'chaine', <, >, |, &, ;,
	ou un mot au \ prs) 
	dans *s et avance *s au debut du prochain token
     */
{
  char *start, *token;

  for (; **s == ' '; (*s)++);

  start = *s;
  if (*start == '\'') {
    (*s)++;
    for (; **s && **s != '\''; (*s)++);
    if (**s != '\'') {
      fprintf (stderr, "Syntax error: ' missing.\n");
      return NULL;
    }
    (*s)++;
  } else if (*start == '"') {
    (*s)++;
    for (; **s && **s != '"' && **s != '\''; (*s)++)
      if (**s == '\\') (*s)++;
    if (**s != '"') {
      fprintf (stderr, "Syntax error: \" missing.\n");
      return NULL;
    }
    (*s)++;
  } else if (strchr ("<>|&;", *start)) {
    *s = start + 1;
  } else {
    for (; **s && (!strchr (" '\"<>|&;", **s));	(*s)++)
      if (**s == '\\') (*s)++;
  }
  
  token = malloc (*s - start + 1);
  strncpy (token, start, (int) (*s - start));
  token[*s - start] = '\0';

  for (; **s == ' '; (*s)++);

  return token;
}


token_list_t get_token_list (char *s)
{
  char *sp, *token;
  token_list_t l, *p;

  sp = s;
  p = &l;
  while (*sp) {
    token = take (&sp);
    if (!token) {
      *p = NULL;
      free_token_list (l);
      return NULL;
    }
    if (!strcmp (token, "")) break;
    *p = malloc (sizeof **p);
    (*p)->token = token;
    (*p)->used = 0;
    p = &((*p)->next);
  }
  *p = NULL;
  return l;
}

/* Toutes les fonctions get_xxx suivantes (sauf get_comm) renvoient le prochain
   xxx dans la token_list l (si il y en a un) et avance l sur le token suivant.
   si il n'y a pas de xxx en tete de l, elles renvoient NULL et ne touchent pas a l
 */

char *get_word (token_list_t *l)
{
  char *word;

  if (!*l) return NULL;
  
  word = (*l)->token;
  if (strcmp (word, "<") &&
      strcmp (word, ">") &&
      strcmp (word, "|") &&
      strcmp (word, ";") &&
      strcmp (word, "&")) {
    (*l)->used = 1;
    *l = (*l)->next;
    return word;
  }
  return NULL;
}


elt_t get_elt (token_list_t *l)
{
  token_list_t p = *l;
  elt_t elt = {ARG, NULL};

  if ((elt.word = get_word (l)))
    return elt;
  else if (!strcmp (p->token, ">")) {
    elt.type = OUT;
    *l = (*l)->next;
    elt.word = get_word (l);
  } else if (!strcmp (p->token, "<")) {
    elt.type = IN;
    *l = (*l)->next;
    elt.word = get_word (l);
  }
  if (!elt.word)
    *l = p;
  return elt;
}

elt_list_t get_elt_list (token_list_t *l)
{
  elt_list_t elt_list, *p;
  elt_t elt;

  p = &elt_list;
  while (*l) {
    elt = get_elt (l);
    if (!(elt.word)) break;
    *p = malloc (sizeof **p);
    (*p)->elt = elt;
    p = &((*p)->next);
  }
  *p = NULL;
  return elt_list;
}

scomm_t get_scomm (token_list_t *l)
{
  scomm_t scomm = {NULL, NULL};

  scomm.name = get_word (l);
  if (!scomm.name) return scomm;
  scomm.elt_list = get_elt_list (l);
  return scomm;
}

pipe_t get_pipe (token_list_t *l)
{
  pipe_t pipe, *p;
  scomm_t scomm;

  p = &pipe;
  while (*l) {
    scomm = get_scomm (l);
    if (!(scomm.name)) break;
    *p = malloc (sizeof **p);
    (*p)->scomm = scomm;
    (*p)->pid = (pid_t) -1;
    p = &((*p)->next);
    if ((!(*l)) || (strcmp ((*l)->token, "|")))
      break;
    else
      *l = (*l)->next;
  }
  *p = NULL;
  return pipe;
}

comm_t get_comm (token_list_t l)
     /*  renvoie la commande contenue dans l, 
	 renvoie NULL et affiche un message
	 si erreur de syntaxe
      */
{
  token_list_t tok = l;
  comm_t comm = NULL, *p;
  pipe_t pipe;

  p = &comm;
  while (tok) {
    pipe = get_pipe (&tok);
    *p = malloc (sizeof **p);
    (*p)->pipe = pipe;
    if (!tok || !strcmp(tok->token, ";"))
      (*p)->sync = 1;
    else if (!strcmp(tok->token, "&"))
      (*p)->sync = 0;
    else {
      (*p)->next = NULL;
      _free_comm (comm, SOFT);
      fprintf (stderr, "Parse error near %s\n",tok->token);
      return NULL;
    }
    if (tok)
      tok = tok->next;
    p = &((*p)->next);
  }
  *p = NULL;
  return comm;  
}



// --- POUR LES TESTS


void print_token_list (token_list_t l)
{
  token_list_t p = l;

  while (p) {
    printf ("'%s' ", p->token);
    p = p->next;
  }
  printf ("\n");
}

void print_elt_list (elt_list_t l)
{
  elt_list_t p = l;

  while (p) {
    switch (p->elt.type) {
    case ARG: printf ("<");break;
    case  IN: printf ("<in: "); break;
    case OUT: printf ("<out: "); break;
    }
    printf ("%s> ", p->elt.word);
    p = p->next;
  }
  printf ("\n");
}

void print_scomm (scomm_t scomm)
{
  printf ("  - %s, ", scomm.name);
  print_elt_list (scomm.elt_list);
}

void print_pipe (pipe_t l)
{
  pipe_t p = l;

  while (p) {
    print_scomm (p->scomm);
    p = p->next;
  }
  printf ("\n");
}

void print_comm (comm_t comm)
{
  while (comm) {
    if (comm->sync)
      printf ("synchrone :\n");
    else 
      printf ("asynchrone :\n");
    print_pipe (comm->pipe);
    comm = comm->next;
  }
}


// Fonction principale du module

comm_t read_comm (FILE *f)
{
  char line[MAX_LINE_SIZE];
  char *exit_argv[2] = {"0", NULL};
  int r;
  token_list_t l;
  comm_t comm;

  r = read_line (line, f);
  if (r == E_FGETS) {
    printf ("\n");
    update_print_if_dead_jobs (stdout);
    dd_exit (stdin, stdout, exit_argv);
  } else if (r == E_TOO_LONG) {
    fprintf (stderr, "Error: line too long.\n");
    return NULL;
  } 
  l = get_token_list (line);
  comm = get_comm (l);
  free_token_list (l);

  return comm;
}
